@@ -55,6 +55,9 @@ const _createTrustedTypesPolicy = function (trustedTypes, document) {
       createHTML(html) {
         return html;
       },
+      createScriptURL(scriptUrl) {
+        return scriptUrl;
+      },
     });
   } catch (_) {
     // Policy creation failed (most likely another DOMPurify script has
@@ -100,8 +103,7 @@ function createDOMPurify(window = getGlobal()) {
     Element,
     NodeFilter,
     NamedNodeMap = window.NamedNodeMap || window.MozNamedAttrMap,
-    Text,
-    Comment,
+    HTMLFormElement,
     DOMParser,
     trustedTypes,
   } = window;
@@ -130,10 +132,7 @@ function createDOMPurify(window = getGlobal()) {
     trustedTypes,
     originalDocument
   );
-  const emptyHTML =
-    trustedTypesPolicy && RETURN_TRUSTED_TYPE
-      ? trustedTypesPolicy.createHTML('')
-      : '';
+  const emptyHTML = trustedTypesPolicy ? trustedTypesPolicy.createHTML('') : '';
 
   const {
     implementation,
@@ -194,6 +193,35 @@ function createDOMPurify(window = getGlobal()) {
     ...ATTRS.xml,
   ]);
 
+  /*
+   * Configure how DOMPUrify should handle custom elements and their attributes as well as customized built-in elements.
+   * @property {RegExp|Function|null} tagNameCheck one of [null, regexPattern, predicate]. Default: `null` (disallow any custom elements)
+   * @property {RegExp|Function|null} attributeNameCheck one of [null, regexPattern, predicate]. Default: `null` (disallow any attributes not on the allow list)
+   * @property {boolean} allowCustomizedBuiltInElements allow custom elements derived from built-ins if they pass CUSTOM_ELEMENT_HANDLING.tagNameCheck. Default: `false`.
+   */
+  const CUSTOM_ELEMENT_HANDLING = Object.seal(
+    Object.create(null, {
+      tagNameCheck: {
+        writable: true,
+        configurable: false,
+        enumerable: true,
+        value: null,
+      },
+      attributeNameCheck: {
+        writable: true,
+        configurable: false,
+        enumerable: true,
+        value: null,
+      },
+      allowCustomizedBuiltInElements: {
+        writable: true,
+        configurable: false,
+        enumerable: true,
+        value: false,
+      },
+    })
+  );
+
   /* Explicitly forbidden tags (overrides ALLOWED_TAGS/ADD_TAGS) */
   let FORBID_TAGS = null;
 
@@ -234,24 +262,31 @@ function createDOMPurify(window = getGlobal()) {
    * string  (or a TrustedHTML object if Trusted Types are supported) */
   let RETURN_DOM_FRAGMENT = false;
 
-  /* If `RETURN_DOM` or `RETURN_DOM_FRAGMENT` is enabled, decide if the returned DOM
-   * `Node` is imported into the current `Document`. If this flag is not enabled the
-   * `Node` will belong (its ownerDocument) to a fresh `HTMLDocument`, created by
-   * DOMPurify.
-   *
-   * This defaults to `true` starting DOMPurify 2.2.0. Note that setting it to `false`
-   * might cause XSS from attacks hidden in closed shadowroots in case the browser
-   * supports Declarative Shadow: DOM https://web.dev/declarative-shadow-dom/
-   */
-  let RETURN_DOM_IMPORT = true;
-
   /* Try to return a Trusted Type object instead of a string, return a string in
    * case Trusted Types are not supported  */
   let RETURN_TRUSTED_TYPE = false;
 
-  /* Output should be free from DOM clobbering attacks? */
+  /* Output should be free from DOM clobbering attacks?
+   * This sanitizes markups named with colliding, clobberable built-in DOM APIs.
+   */
   let SANITIZE_DOM = true;
 
+  /* Achieve full DOM Clobbering protection by isolating the namespace of named
+   * properties and JS variables, mitigating attacks that abuse the HTML/DOM spec rules.
+   *
+   * HTML/DOM spec rules that enable DOM Clobbering:
+   *   - Named Access on Window (§7.3.3)
+   *   - DOM Tree Accessors (§3.1.5)
+   *   - Form Element Parent-Child Relations (§4.10.3)
+   *   - Iframe srcdoc / Nested WindowProxies (§4.8.5)
+   *   - HTMLCollection (§4.2.10.2)
+   *
+   * Namespace isolation is implemented by prefixing `id` and `name` attributes
+   * with a constant string, i.e., `user-content-`
+   */
+  let SANITIZE_NAMED_PROPS = false;
+  const SANITIZE_NAMED_PROPS_PREFIX = 'user-content-';
+
   /* Keep element content when removing element? */
   let KEEP_CONTENT = true;
 
@@ -343,6 +378,10 @@ function createDOMPurify(window = getGlobal()) {
 
   const formElement = document.createElement('form');
 
+  const isRegexOrFunction = function (testValue) {
+    return testValue instanceof RegExp || testValue instanceof Function;
+  };
+
   /**
    * _parseConfig
    *
@@ -362,46 +401,6 @@ function createDOMPurify(window = getGlobal()) {
     /* Shield configuration object from prototype pollution */
     cfg = clone(cfg);
 
-    /* Set configuration parameters */
-    ALLOWED_TAGS =
-      'ALLOWED_TAGS' in cfg
-        ? addToSet({}, cfg.ALLOWED_TAGS)
-        : DEFAULT_ALLOWED_TAGS;
-    ALLOWED_ATTR =
-      'ALLOWED_ATTR' in cfg
-        ? addToSet({}, cfg.ALLOWED_ATTR)
-        : DEFAULT_ALLOWED_ATTR;
-    URI_SAFE_ATTRIBUTES =
-      'ADD_URI_SAFE_ATTR' in cfg
-        ? addToSet(clone(DEFAULT_URI_SAFE_ATTRIBUTES), cfg.ADD_URI_SAFE_ATTR)
-        : DEFAULT_URI_SAFE_ATTRIBUTES;
-    DATA_URI_TAGS =
-      'ADD_DATA_URI_TAGS' in cfg
-        ? addToSet(clone(DEFAULT_DATA_URI_TAGS), cfg.ADD_DATA_URI_TAGS)
-        : DEFAULT_DATA_URI_TAGS;
-    FORBID_CONTENTS =
-      'FORBID_CONTENTS' in cfg
-        ? addToSet({}, cfg.FORBID_CONTENTS)
-        : DEFAULT_FORBID_CONTENTS;
-    FORBID_TAGS = 'FORBID_TAGS' in cfg ? addToSet({}, cfg.FORBID_TAGS) : {};
-    FORBID_ATTR = 'FORBID_ATTR' in cfg ? addToSet({}, cfg.FORBID_ATTR) : {};
-    USE_PROFILES = 'USE_PROFILES' in cfg ? cfg.USE_PROFILES : false;
-    ALLOW_ARIA_ATTR = cfg.ALLOW_ARIA_ATTR !== false; // Default true
-    ALLOW_DATA_ATTR = cfg.ALLOW_DATA_ATTR !== false; // Default true
-    ALLOW_UNKNOWN_PROTOCOLS = cfg.ALLOW_UNKNOWN_PROTOCOLS || false; // Default false
-    SAFE_FOR_TEMPLATES = cfg.SAFE_FOR_TEMPLATES || false; // Default false
-    WHOLE_DOCUMENT = cfg.WHOLE_DOCUMENT || false; // Default false
-    RETURN_DOM = cfg.RETURN_DOM || false; // Default false
-    RETURN_DOM_FRAGMENT = cfg.RETURN_DOM_FRAGMENT || false; // Default false
-    RETURN_DOM_IMPORT = cfg.RETURN_DOM_IMPORT !== false; // Default true
-    RETURN_TRUSTED_TYPE = cfg.RETURN_TRUSTED_TYPE || false; // Default false
-    FORCE_BODY = cfg.FORCE_BODY || false; // Default false
-    SANITIZE_DOM = cfg.SANITIZE_DOM !== false; // Default true
-    KEEP_CONTENT = cfg.KEEP_CONTENT !== false; // Default true
-    IN_PLACE = cfg.IN_PLACE || false; // Default false
-    IS_ALLOWED_URI = cfg.ALLOWED_URI_REGEXP || IS_ALLOWED_URI;
-    NAMESPACE = cfg.NAMESPACE || HTML_NAMESPACE;
-
     PARSER_MEDIA_TYPE =
       // eslint-disable-next-line unicorn/prefer-includes
       SUPPORTED_PARSER_MEDIA_TYPES.indexOf(cfg.PARSER_MEDIA_TYPE) === -1
@@ -414,6 +413,84 @@ function createDOMPurify(window = getGlobal()) {
         ? (x) => x
         : stringToLowerCase;
 
+    /* Set configuration parameters */
+    ALLOWED_TAGS =
+      'ALLOWED_TAGS' in cfg
+        ? addToSet({}, cfg.ALLOWED_TAGS, transformCaseFunc)
+        : DEFAULT_ALLOWED_TAGS;
+    ALLOWED_ATTR =
+      'ALLOWED_ATTR' in cfg
+        ? addToSet({}, cfg.ALLOWED_ATTR, transformCaseFunc)
+        : DEFAULT_ALLOWED_ATTR;
+    URI_SAFE_ATTRIBUTES =
+      'ADD_URI_SAFE_ATTR' in cfg
+        ? addToSet(
+            clone(DEFAULT_URI_SAFE_ATTRIBUTES), // eslint-disable-line indent
+            cfg.ADD_URI_SAFE_ATTR, // eslint-disable-line indent
+            transformCaseFunc // eslint-disable-line indent
+          ) // eslint-disable-line indent
+        : DEFAULT_URI_SAFE_ATTRIBUTES;
+    DATA_URI_TAGS =
+      'ADD_DATA_URI_TAGS' in cfg
+        ? addToSet(
+            clone(DEFAULT_DATA_URI_TAGS), // eslint-disable-line indent
+            cfg.ADD_DATA_URI_TAGS, // eslint-disable-line indent
+            transformCaseFunc // eslint-disable-line indent
+          ) // eslint-disable-line indent
+        : DEFAULT_DATA_URI_TAGS;
+    FORBID_CONTENTS =
+      'FORBID_CONTENTS' in cfg
+        ? addToSet({}, cfg.FORBID_CONTENTS, transformCaseFunc)
+        : DEFAULT_FORBID_CONTENTS;
+    FORBID_TAGS =
+      'FORBID_TAGS' in cfg
+        ? addToSet({}, cfg.FORBID_TAGS, transformCaseFunc)
+        : {};
+    FORBID_ATTR =
+      'FORBID_ATTR' in cfg
+        ? addToSet({}, cfg.FORBID_ATTR, transformCaseFunc)
+        : {};
+    USE_PROFILES = 'USE_PROFILES' in cfg ? cfg.USE_PROFILES : false;
+    ALLOW_ARIA_ATTR = cfg.ALLOW_ARIA_ATTR !== false; // Default true
+    ALLOW_DATA_ATTR = cfg.ALLOW_DATA_ATTR !== false; // Default true
+    ALLOW_UNKNOWN_PROTOCOLS = cfg.ALLOW_UNKNOWN_PROTOCOLS || false; // Default false
+    SAFE_FOR_TEMPLATES = cfg.SAFE_FOR_TEMPLATES || false; // Default false
+    WHOLE_DOCUMENT = cfg.WHOLE_DOCUMENT || false; // Default false
+    RETURN_DOM = cfg.RETURN_DOM || false; // Default false
+    RETURN_DOM_FRAGMENT = cfg.RETURN_DOM_FRAGMENT || false; // Default false
+    RETURN_TRUSTED_TYPE = cfg.RETURN_TRUSTED_TYPE || false; // Default false
+    FORCE_BODY = cfg.FORCE_BODY || false; // Default false
+    SANITIZE_DOM = cfg.SANITIZE_DOM !== false; // Default true
+    SANITIZE_NAMED_PROPS = cfg.SANITIZE_NAMED_PROPS || false; // Default false
+    KEEP_CONTENT = cfg.KEEP_CONTENT !== false; // Default true
+    IN_PLACE = cfg.IN_PLACE || false; // Default false
+    IS_ALLOWED_URI = cfg.ALLOWED_URI_REGEXP || IS_ALLOWED_URI;
+    NAMESPACE = cfg.NAMESPACE || HTML_NAMESPACE;
+    if (
+      cfg.CUSTOM_ELEMENT_HANDLING &&
+      isRegexOrFunction(cfg.CUSTOM_ELEMENT_HANDLING.tagNameCheck)
+    ) {
+      CUSTOM_ELEMENT_HANDLING.tagNameCheck =
+        cfg.CUSTOM_ELEMENT_HANDLING.tagNameCheck;
+    }
+
+    if (
+      cfg.CUSTOM_ELEMENT_HANDLING &&
+      isRegexOrFunction(cfg.CUSTOM_ELEMENT_HANDLING.attributeNameCheck)
+    ) {
+      CUSTOM_ELEMENT_HANDLING.attributeNameCheck =
+        cfg.CUSTOM_ELEMENT_HANDLING.attributeNameCheck;
+    }
+
+    if (
+      cfg.CUSTOM_ELEMENT_HANDLING &&
+      typeof cfg.CUSTOM_ELEMENT_HANDLING.allowCustomizedBuiltInElements ===
+        'boolean'
+    ) {
+      CUSTOM_ELEMENT_HANDLING.allowCustomizedBuiltInElements =
+        cfg.CUSTOM_ELEMENT_HANDLING.allowCustomizedBuiltInElements;
+    }
+
     if (SAFE_FOR_TEMPLATES) {
       ALLOW_DATA_ATTR = false;
     }
@@ -456,7 +533,7 @@ function createDOMPurify(window = getGlobal()) {
         ALLOWED_TAGS = clone(ALLOWED_TAGS);
       }
 
-      addToSet(ALLOWED_TAGS, cfg.ADD_TAGS);
+      addToSet(ALLOWED_TAGS, cfg.ADD_TAGS, transformCaseFunc);
     }
 
     if (cfg.ADD_ATTR) {
@@ -464,11 +541,11 @@ function createDOMPurify(window = getGlobal()) {
         ALLOWED_ATTR = clone(ALLOWED_ATTR);
       }
 
-      addToSet(ALLOWED_ATTR, cfg.ADD_ATTR);
+      addToSet(ALLOWED_ATTR, cfg.ADD_ATTR, transformCaseFunc);
     }
 
     if (cfg.ADD_URI_SAFE_ATTR) {
-      addToSet(URI_SAFE_ATTRIBUTES, cfg.ADD_URI_SAFE_ATTR);
+      addToSet(URI_SAFE_ATTRIBUTES, cfg.ADD_URI_SAFE_ATTR, transformCaseFunc);
     }
 
     if (cfg.FORBID_CONTENTS) {
@@ -476,7 +553,7 @@ function createDOMPurify(window = getGlobal()) {
         FORBID_CONTENTS = clone(FORBID_CONTENTS);
       }
 
-      addToSet(FORBID_CONTENTS, cfg.FORBID_CONTENTS);
+      addToSet(FORBID_CONTENTS, cfg.FORBID_CONTENTS, transformCaseFunc);
     }
 
     /* Add #text in case KEEP_CONTENT is set to true */
@@ -519,6 +596,18 @@ function createDOMPurify(window = getGlobal()) {
     'annotation-xml',
   ]);
 
+  // Certain elements are allowed in both SVG and HTML
+  // namespace. We need to specify them explicitly
+  // so that they don't get erroneously deleted from
+  // HTML namespace.
+  const COMMON_SVG_AND_HTML_ELEMENTS = addToSet({}, [
+    'title',
+    'style',
+    'font',
+    'a',
+    'script',
+  ]);
+
   /* Keep track of all possible SVG and MathML tags
    * so that we can perform the namespace checks
    * correctly. */
@@ -613,23 +702,11 @@ function createDOMPurify(window = getGlobal()) {
         return false;
       }
 
-      // Certain elements are allowed in both SVG and HTML
-      // namespace. We need to specify them explicitly
-      // so that they don't get erronously deleted from
-      // HTML namespace.
-      const commonSvgAndHTMLElements = addToSet({}, [
-        'title',
-        'style',
-        'font',
-        'a',
-        'script',
-      ]);
-
       // We disallow tags that are specific for MathML
       // or SVG and should never appear in HTML namespace
       return (
         !ALL_MATHML_TAGS[tagName] &&
-        (commonSvgAndHTMLElements[tagName] || !ALL_SVG_TAGS[tagName])
+        (COMMON_SVG_AND_HTML_ELEMENTS[tagName] || !ALL_SVG_TAGS[tagName])
       );
     }
 
@@ -773,6 +850,7 @@ function createDOMPurify(window = getGlobal()) {
     return createNodeIterator.call(
       root.ownerDocument || root,
       root,
+      // eslint-disable-next-line no-bitwise
       NodeFilter.SHOW_ELEMENT | NodeFilter.SHOW_COMMENT | NodeFilter.SHOW_TEXT,
       null,
       false
@@ -786,24 +864,17 @@ function createDOMPurify(window = getGlobal()) {
    * @return {Boolean} true if clobbered, false if safe
    */
   const _isClobbered = function (elm) {
-    if (elm instanceof Text || elm instanceof Comment) {
-      return false;
-    }
-
-    if (
-      typeof elm.nodeName !== 'string' ||
-      typeof elm.textContent !== 'string' ||
-      typeof elm.removeChild !== 'function' ||
-      !(elm.attributes instanceof NamedNodeMap) ||
-      typeof elm.removeAttribute !== 'function' ||
-      typeof elm.setAttribute !== 'function' ||
-      typeof elm.namespaceURI !== 'string' ||
-      typeof elm.insertBefore !== 'function'
-    ) {
-      return true;
-    }
-
-    return false;
+    return (
+      elm instanceof HTMLFormElement &&
+      (typeof elm.nodeName !== 'string' ||
+        typeof elm.textContent !== 'string' ||
+        typeof elm.removeChild !== 'function' ||
+        !(elm.attributes instanceof NamedNodeMap) ||
+        typeof elm.removeAttribute !== 'function' ||
+        typeof elm.setAttribute !== 'function' ||
+        typeof elm.namespaceURI !== 'string' ||
+        typeof elm.insertBefore !== 'function')
+    );
   };
 
   /**
@@ -862,7 +933,7 @@ function createDOMPurify(window = getGlobal()) {
     }
 
     /* Check if tagname contains Unicode */
-    if (stringMatch(currentNode.nodeName, /[\u0080-\uFFFF]/)) {
+    if (regExpTest(/[\u0080-\uFFFF]/, currentNode.nodeName)) {
       _forceRemove(currentNode);
       return true;
     }
@@ -878,6 +949,7 @@ function createDOMPurify(window = getGlobal()) {
 
     /* Detect mXSS attempts abusing namespace confusion */
     if (
+      currentNode.hasChildNodes() &&
       !_isNode(currentNode.firstElementChild) &&
       (!_isNode(currentNode.content) ||
         !_isNode(currentNode.content.firstElementChild)) &&
@@ -899,6 +971,20 @@ function createDOMPurify(window = getGlobal()) {
 
     /* Remove element if anything forbids its presence */
     if (!ALLOWED_TAGS[tagName] || FORBID_TAGS[tagName]) {
+      /* Check if we have a custom element to handle */
+      if (!FORBID_TAGS[tagName] && _basicCustomElementTest(tagName)) {
+        if (
+          CUSTOM_ELEMENT_HANDLING.tagNameCheck instanceof RegExp &&
+          regExpTest(CUSTOM_ELEMENT_HANDLING.tagNameCheck, tagName)
+        )
+          return false;
+        if (
+          CUSTOM_ELEMENT_HANDLING.tagNameCheck instanceof Function &&
+          CUSTOM_ELEMENT_HANDLING.tagNameCheck(tagName)
+        )
+          return false;
+      }
+
       /* Keep content except for bad-listed elements */
       if (KEEP_CONTENT && !FORBID_CONTENTS[tagName]) {
         const parentNode = getParentNode(currentNode) || currentNode.parentNode;
@@ -985,8 +1071,33 @@ function createDOMPurify(window = getGlobal()) {
       // This attribute is safe
       /* Otherwise, check the name is permitted */
     } else if (!ALLOWED_ATTR[lcName] || FORBID_ATTR[lcName]) {
-      return false;
-
+      if (
+        // First condition does a very basic check if a) it's basically a valid custom element tagname AND
+        // b) if the tagName passes whatever the user has configured for CUSTOM_ELEMENT_HANDLING.tagNameCheck
+        // and c) if the attribute name passes whatever the user has configured for CUSTOM_ELEMENT_HANDLING.attributeNameCheck
+        (_basicCustomElementTest(lcTag) &&
+          ((CUSTOM_ELEMENT_HANDLING.tagNameCheck instanceof RegExp &&
+            regExpTest(CUSTOM_ELEMENT_HANDLING.tagNameCheck, lcTag)) ||
+            (CUSTOM_ELEMENT_HANDLING.tagNameCheck instanceof Function &&
+              CUSTOM_ELEMENT_HANDLING.tagNameCheck(lcTag))) &&
+          ((CUSTOM_ELEMENT_HANDLING.attributeNameCheck instanceof RegExp &&
+            regExpTest(CUSTOM_ELEMENT_HANDLING.attributeNameCheck, lcName)) ||
+            (CUSTOM_ELEMENT_HANDLING.attributeNameCheck instanceof Function &&
+              CUSTOM_ELEMENT_HANDLING.attributeNameCheck(lcName)))) ||
+        // Alternative, second condition checks if it's an `is`-attribute, AND
+        // the value passes whatever the user has configured for CUSTOM_ELEMENT_HANDLING.tagNameCheck
+        (lcName === 'is' &&
+          CUSTOM_ELEMENT_HANDLING.allowCustomizedBuiltInElements &&
+          ((CUSTOM_ELEMENT_HANDLING.tagNameCheck instanceof RegExp &&
+            regExpTest(CUSTOM_ELEMENT_HANDLING.tagNameCheck, value)) ||
+            (CUSTOM_ELEMENT_HANDLING.tagNameCheck instanceof Function &&
+              CUSTOM_ELEMENT_HANDLING.tagNameCheck(value))))
+      ) {
+        // If user has supplied a regexp or function in CUSTOM_ELEMENT_HANDLING.tagNameCheck, we need to also allow derived custom elements using the same tagName test.
+        // Additionally, we need to allow attributes passing the CUSTOM_ELEMENT_HANDLING.attributeNameCheck user has configured, as custom elements can define these at their own discretion.
+      } else {
+        return false;
+      }
       /* Check value is safe. First, is attr inert? If so, is safe */
     } else if (URI_SAFE_ATTRIBUTES[lcName]) {
       // This attribute is safe
@@ -1025,6 +1136,16 @@ function createDOMPurify(window = getGlobal()) {
     return true;
   };
 
+  /**
+   * _basicCustomElementCheck
+   * checks if at least one dash is included in tagName, and it's not the first char
+   * for more sophisticated checking see https://github.com/sindresorhus/validate-element-name
+   * @param {string} tagName name of the tag of the node to sanitize
+   */
+  const _basicCustomElementTest = function (tagName) {
+    return tagName.indexOf('-') > 0;
+  };
+
   /**
    * _sanitizeAttributes
    *
@@ -1062,7 +1183,7 @@ function createDOMPurify(window = getGlobal()) {
     while (l--) {
       attr = attributes[l];
       const { name, namespaceURI } = attr;
-      value = stringTrim(attr.value);
+      value = name === 'value' ? attr.value : stringTrim(attr.value);
       lcName = transformCaseFunc(name);
 
       /* Execute a hook if present */
@@ -1103,6 +1224,39 @@ function createDOMPurify(window = getGlobal()) {
         continue;
       }
 
+      /* Full DOM Clobbering protection via namespace isolation,
+       * Prefix id and name attributes with `user-content-`
+       */
+      if (SANITIZE_NAMED_PROPS && (lcName === 'id' || lcName === 'name')) {
+        // Remove the attribute with this value
+        _removeAttribute(name, currentNode);
+
+        // Prefix the value and later re-create the attribute with the sanitized value
+        value = SANITIZE_NAMED_PROPS_PREFIX + value;
+      }
+
+      /* Handle attributes that require Trusted Types */
+      if (
+        trustedTypesPolicy &&
+        typeof trustedTypes === 'object' &&
+        typeof trustedTypes.getAttributeType === 'function'
+      ) {
+        if (namespaceURI) {
+          /* Namespaces are not yet supported, see https://bugs.chromium.org/p/chromium/issues/detail?id=1305293 */
+        } else {
+          switch (trustedTypes.getAttributeType(lcTag, lcName)) {
+            case 'TrustedHTML':
+              value = trustedTypesPolicy.createHTML(value);
+              break;
+            case 'TrustedScriptURL':
+              value = trustedTypesPolicy.createScriptURL(value);
+              break;
+            default:
+              break;
+          }
+        }
+      }
+
       /* Handle invalid data-* attribute set by try-catching it */
       try {
         if (namespaceURI) {
@@ -1162,7 +1316,7 @@ function createDOMPurify(window = getGlobal()) {
    * @param {Object} configuration object
    */
   // eslint-disable-next-line complexity
-  DOMPurify.sanitize = function (dirty, cfg) {
+  DOMPurify.sanitize = function (dirty, cfg = {}) {
     let body;
     let importedNode;
     let currentNode;
@@ -1221,7 +1375,15 @@ function createDOMPurify(window = getGlobal()) {
     }
 
     if (IN_PLACE) {
-      /* No special handling necessary for in-place sanitization */
+      /* Do some early pre-sanitization to avoid unsafe root nodes */
+      if (dirty.nodeName) {
+        const tagName = transformCaseFunc(dirty.nodeName);
+        if (!ALLOWED_TAGS[tagName] || FORBID_TAGS[tagName]) {
+          throw typeErrorCreate(
+            'root node is forbidden and cannot be sanitized in-place'
+          );
+        }
+      }
     } else if (dirty instanceof Node) {
       /* If dirty is a DOM element, append to an empty document to avoid
          elements being stripped by the parser */
@@ -1255,7 +1417,7 @@ function createDOMPurify(window = getGlobal()) {
 
       /* Check we have a DOM node from the data */
       if (!body) {
-        return RETURN_DOM ? null : emptyHTML;
+        return RETURN_DOM ? null : RETURN_TRUSTED_TYPE ? emptyHTML : '';
       }
     }
 
@@ -1310,7 +1472,7 @@ function createDOMPurify(window = getGlobal()) {
         returnNode = body;
       }
 
-      if (RETURN_DOM_IMPORT) {
+      if (ALLOWED_ATTR.shadowroot) {
         /*
           AdoptNode() is not used because internal state is not reset
           (e.g. the past names map of a HTMLFormElement), this is safe
@@ -1326,6 +1488,19 @@ function createDOMPurify(window = getGlobal()) {
 
     let serializedHTML = WHOLE_DOCUMENT ? body.outerHTML : body.innerHTML;
 
+    /* Serialize doctype if allowed */
+    if (
+      WHOLE_DOCUMENT &&
+      ALLOWED_TAGS['!doctype'] &&
+      body.ownerDocument &&
+      body.ownerDocument.doctype &&
+      body.ownerDocument.doctype.name &&
+      regExpTest(EXPRESSIONS.DOCTYPE_NAME, body.ownerDocument.doctype.name)
+    ) {
+      serializedHTML =
+        '<!DOCTYPE ' + body.ownerDocument.doctype.name + '>\n' + serializedHTML;
+    }
+
     /* Sanitize final string template-safe */
     if (SAFE_FOR_TEMPLATES) {
       serializedHTML = stringReplace(serializedHTML, MUSTACHE_EXPR, ' ');
@@ -1401,10 +1576,11 @@ function createDOMPurify(window = getGlobal()) {
    * (pops it from the stack of hooks if more are present)
    *
    * @param {String} entryPoint entry point for the hook to remove
+   * @return {Function} removed(popped) hook
    */
   DOMPurify.removeHook = function (entryPoint) {
     if (hooks[entryPoint]) {
-      arrayPop(hooks[entryPoint]);
+      return arrayPop(hooks[entryPoint]);
     }
   };
 
